/*!
 * Copyright (C) 2002 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *
 *    This product includes software developed by egnite Software GmbH
 *    and its contributors.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 */

/*!
 * $Log: relays.c,v $
 * Revision 1.1  2002/06/12 11:32:54  harald
 * First check in
 *
 */

#include <sys/thread.h>
#include <sys/timer.h>
#include <sys/print.h>

#include <dev/spidigio.h>

/*!
 * \example relays/relays.c
 *
 * This sample demonstrates how to control the SPI Octal
 * Relay Output Board.
 */


/*! \fn NutMain(void *arg)
 * \brief Main application routine. 
 *
 * Nut/OS automatically calls this entry after initialization.
 */
THREAD(NutMain, arg)
{
    u_long oval = 1;
    u_long ival;
    u_char ni;
    u_char no;
    u_long baud = 115200;
    
    /*
     * Initialize debug UART device and
     * print banner.
     */
    outp(BV(RXEN) | BV(TXEN), UCR);
    NutDeviceIOCtl(0, UART_SETSPEED, &baud);
    NutPrintString_P(0, PSTR("\r\nDigital Relay Output Demo\r\n"));
    
    /*
     * The outer endless loop is used to keep looking
     * for outputs.
     */
    for(;;) {
        /*
         * Initialize the SPI to the digital I/O boards.
         * The call returns the number of inputs and
         * outputs detected.
         */
        SpiDigitalInit(&ni, &no);
        NutPrintFormat(0, "Found %u digital inputs and %u digital outputs\r\n", ni, no);
        NutSleep(500);

        if(no) {            
            /*
             * The inner endless loop switches
             * a single relays.
             */
            for(;;) {
                /*
                 * Set outputs.
                 */
                SpiDigitalSet(no, oval);
                
                /*
                 * Get inputs and print the result.
                 */
                ival = SpiDigitalGet(ni + no);
                NutPrintFormat(0, "Out: %08lX - In: %08lX\r\n", oval, ival);
                
                /*
                 * Shift around the walking bit value.
                 */
                if((oval <<= 1) == BV(no))
                    oval = 1;
                NutSleep(500);
            }
        }
    }
}
