/*!
 * Copyright (C) 2001-2002 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *
 *    This product includes software developed by egnite Software GmbH
 *    and its contributors.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 */

/*!
 * $Log: threads.c,v $
 * Revision 1.5  2002/06/26 17:29:06  harald
 * First pre-release with 2.4 stack
 *
 * Revision 1.4  2002/06/04 19:13:06  harald
 * *** empty log message ***
 *
 * Revision 1.3  2002/05/08 16:02:33  harald
 * First Imagecraft compilation
 *
 * Revision 1.2  2001/12/04 16:45:40  harald
 * Stack space increased
 *
 * Revision 1.1  2001/06/28 18:43:13  harald
 * Preview release
 *
 */

/*!
 * \example threads/threads.c
 *
 * This sample demonstrates Nut/OS multithreading.
 */

#include <dev/uartavr.h>
#include <sys/thread.h>
#include <sys/print.h>
#include <sys/timer.h>

static NUTDEVICE *uart;

/*
 * High priority thread.
 */
THREAD(Thread1, arg)
{
    /*
     * Endless loop in high priority thread.
     */
    NutThreadSetPriority(16);
    for(;;) {
        NutPrintString(uart, "H");
        NutPrintFlush(uart);
        NutSleep(200);
    }
}

/*
 * Low priority thread.
 */
THREAD(Thread2, arg)
{
    /*
     * Endless loop in low priority thread.
     */
    NutThreadSetPriority(128);
    for(;;) {
        NutPrintString(uart, "L");
        NutPrintFlush(uart);
        NutSleep(1000);
    }
}

/*
 * Main application thread. 
 */
THREAD(NutMain, arg)
{
    u_long baud = 115200;

    /*
     * Register the UART device, open it and
     * set the baudrate.
     */
    NutRegisterDevice(&devUart0, 0, 0);
    uart = NutDeviceOpen("uart0");
    NutDeviceIOCtl(uart, UART_SETSPEED, &baud);

    NutPrintString_P(uart, PSTR("TimerTest\r\n"));
    NutPrintFlush(uart);

    /*
     * Start two additional threads. All threads
     * are started with priority 64.
     */
    NutThreadCreate("t1", Thread1, 0, 1024);
    NutThreadCreate("t2", Thread2, 0, 1024);

    /*
     * Endless loop in main thread.
     */
    for(;;) {
        NutPrintString(uart, "M");
        NutPrintFlush(uart);
        NutSleep(500);
    }
}

