/*
 * Copyright (C) 2001-2002 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *
 *    This product includes software developed by egnite Software GmbH
 *    and its contributors.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 *
 * -
 * Portions Copyright (C) 2001 Jesper Hansen <jesperh@telia.com>.
 * 
 * This file is part of the yampp system.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/*
 * $Log: vs1001k.c,v $
 * Revision 1.3  2002/06/26 17:29:08  harald
 * First pre-release with 2.4 stack
 *
 */

#ifndef __IMAGECRAFT__
#include <interrupt.h>
#endif

#include <sys/event.h>
#include <sys/timer.h>

#include <dev/vs1001k.h>

static u_char writeDelay = 0;
u_char * volatile tx_ptr;
u_char * volatile wr_ptr;
static volatile u_char tx_act;

static u_char *mem_start;
static u_char *mem_end;
static u_short mem_size;

static HANDLE q_dreq = 0;

/*!
 * Data request interrupt service.
 */
static void VsDataRequest(void *arg)
{
    u_char yd;
    
    if(tx_act) {
        sbi(VS_XCS_PORT, VS_XCS_BIT);

        /*
         * Write MP3 data until either no more
         * data is available or the chip clears
         * the DREQ line.
         */
        for(;;) {

            sbi(VS_BSYNC_PORT, VS_BSYNC_BIT);
            outp(*tx_ptr, SPDR);
#ifdef __IMAGECRAFT__
            asm("nop\nnop\nnop\n");
#else
            asm volatile("nop\n\tnop\n\tnop");
#endif
            cbi(VS_BSYNC_PORT, VS_BSYNC_BIT);

            if(++tx_ptr > mem_end)
                tx_ptr = mem_start;
            loop_until_bit_is_set(SPSR, SPIF);

            /*
             * Stop transfer if our data buffer is empty.
             */
            if(tx_ptr == wr_ptr) {
                tx_act = 0;
                break;
            }

            /*
             * Stop transfer if the chip is filled up.
             */
            if(bit_is_clear(VS_DREQ_PIN, VS_DREQ_BIT)) 
                break;
        }

        /*
         * If the transfer stopped because the decoder
         * clears the DREQ signal, it still has room
         * for 32 bytes. 
         */
        if(tx_act) {
            for(yd = 32; yd; yd--) {

                sbi(VS_BSYNC_PORT, VS_BSYNC_BIT);
                outp(*tx_ptr, SPDR);
#ifdef __IMAGECRAFT__
                asm("nop\nnop\nnop\n");
#else
                asm volatile("nop\n\tnop\n\tnop");
#endif
                cbi(VS_BSYNC_PORT, VS_BSYNC_BIT);

                if(++tx_ptr > mem_end)
                    tx_ptr = mem_start;
                loop_until_bit_is_set(SPSR, SPIF); 

                /*
                 * Stop transfer if our data buffer is empty.
                 */
                if(tx_ptr == wr_ptr) {
                    tx_act = 0;
                    break;
                }
            }
        }
    }
    NutEventPostAsync(&q_dreq);
}

/*!
 * \brief Write a byte to the serial control interface.
 */
#ifdef __IMAGECRAFT__
static void VsSciPutByte(u_char data)
#else
static inline void VsSciPutByte(u_char data)
#endif
{
    u_char mask = 0x80;

    /*
     * Loop until all 8 bits are processed.
     */
    while(mask) {

        /*
         * Set data line.
         */
        if(data & mask)
            sbi(VS_SI_PORT, VS_SI_BIT);
        else
            cbi(VS_SI_PORT, VS_SI_BIT);

        /*
         * Toggle clock and shift mask.
         */
        sbi(VS_SCK_PORT, VS_SCK_BIT);
        mask >>= 1;
        cbi(VS_SCK_PORT, VS_SCK_BIT);
    }
}   

/*
 * Write a word to the VS1001 command interface.
 */
static void VsWriteReg(u_char reg, u_short data)
{
    /*
     * Disable interrupts and select chip.
     */
    cli();
    cbi(VS_XCS_PORT, VS_XCS_BIT);

    cbi(SPCR, SPE);     
    //outp(BV(MSTR) | BV(SPE), SPCR);

    VsSciPutByte(VS_OPCODE_WRITE);
    VsSciPutByte(reg);
    VsSciPutByte((u_char)(data >> 8));
    VsSciPutByte((u_char)data);

    /*
     * Added on Jesper's recommendation.
     */
    if(writeDelay)
        NutDelay(writeDelay);

    /* 
     * Re-enable SPI. Changed due to a hint by Jesper.
     */
    outp(BV(MSTR) | BV(SPE), SPCR);
    outp(inp(SPSR), SPSR);

    /*
     * Deselect chip and enable interrupts.
     */
    sbi(VS_XCS_PORT, VS_XCS_BIT);
    sei();
}

/*
 * Write a byte to the VS1001 data interface.
 */
#ifdef __IMAGECRAFT__
static void VsSend(u_char b)
#else
static inline void VsSend(u_char b)
#endif
{
    /*
     * Set BSYNC high.
     */
    sbi(VS_BSYNC_PORT, VS_BSYNC_BIT);
    outp(b, SPDR);
#ifdef __IMAGECRAFT__
    asm("nop\nnop\nnop\nnop\nnop\nnop\n");
#else
    asm volatile("nop\n\tnop\n\tnop\n\tnop\n\tnop\n\tnop");
#endif

    /*
     * Set BSYNC back to low.
     */
    cbi(VS_BSYNC_PORT, VS_BSYNC_BIT);

    loop_until_bit_is_set(SPSR, SPIF);
}

/*
 * VS1001 software reset.
 */
void VsReset(u_short mode)
{
    int i;

    /*
     * Software reset.
     */
    VsWriteReg(VS_MODE_REG, VS_SM_RESET);
    NutDelay(2);

    /*
     * Wait for DREQ.
     */
    loop_until_bit_is_set(VS_DREQ_PIN, VS_DREQ_BIT);

    NutDelay(200);
    VsWriteReg(VS_CLOCKF_REG, 0x9800);

    for(i = 0; i < 1024; i++)
        VsSend(0);

    /*
     * Set configured modes
     */
    VsWriteReg(VS_MODE_REG, mode);
}

/*
 * Set volume.
 */
void VsSetVolume(u_char left, u_char right)
{
    VsWriteReg(VS_VOL_REG, (((u_short)left) << 8) | (u_short)right);
}

/*
 * Sine wave beep.
 */
void VsBeep(u_char fsin, u_char ms)
{
    u_char i;
    u_char on[] = { 0x53, 0xEF, 0x6E, 56 };
    u_char off[] = { 0x45, 0x78, 0x69, 0x74 };
    
    on[3] += (fsin & 7) * 9;
    for(i = 0; i < sizeof(on); i++)
        VsSend(on[i]);
    for(i = 0; i < 8; i++)
        VsSend(0);
    NutDelay(ms);
    for(i = 0; i < sizeof(off); i++)
        VsSend(off[i]);
    for(i = 0; i < 8; i++)
        VsSend(0);
}


/*
 * Initialize the VS1001 hardware interface.
 */
void VsInit(u_char wDelay)
{
    writeDelay = wDelay;

    /*
     * Toggle MP3 hardware reset output.
     */
    cbi(VS_RESET_PORT, VS_RESET_BIT);
    sbi(VS_RESET_DDR, VS_RESET_BIT);
    NutDelay(3);
    sbi(VS_RESET_PORT, VS_RESET_BIT);

    /*
     * Set BSYNC output low.
     */
    cbi(VS_BSYNC_PORT, VS_BSYNC_BIT);
    sbi(VS_BSYNC_DDR, VS_BSYNC_BIT);

    /*
     * Set MP3 chip select output low.
     */
    sbi(VS_XCS_PORT, VS_XCS_BIT);
    sbi(VS_XCS_DDR, VS_XCS_BIT);

    /*
     * Set DREQ input with pullup.
     */
    cbi(VS_DREQ_DDR, VS_DREQ_BIT);
    sbi(VS_DREQ_PORT, VS_DREQ_BIT);

    // setup serial data interface :
    // clock = f/4
    // select clock phase positive going in middle of data
    // master mode
    // enable SPI

    // setup serial data I/O pins

    /*
     * Set MOSI output.
     * FIXME: hi/lo?
     */
    sbi(VS_SI_DDR, VS_SI_BIT);

    /*
     * Set SS to output for SPI master mode.
     */
    sbi(VS_SS_DDR, VS_SS_BIT);

    /*
     * Set MISO to input.
     */
    cbi(VS_SO_DDR, VS_SO_BIT);

    /*
     * Set SCK output low.
     */
    sbi(VS_SCK_DDR, VS_SCK_BIT);
    cbi(VS_SCK_PORT, VS_SCK_BIT);

    outp(BV(MSTR) | BV(SPE), SPCR);
    inp(SPSR);  // clear status
}

