/*
 * Copyright (C) 2001-2002 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *
 *    This product includes software developed by egnite Software GmbH
 *    and its contributors.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 *
 * -
 * Portions Copyright (C) 2000 David J. Hudson <dave@humbug.demon.co.uk>
 *
 * This file is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You can redistribute this file and/or modify it under the terms of the GNU
 * General Public License (GPL) as published by the Free Software Foundation;
 * either version 2 of the License, or (at your discretion) any later version.
 * See the accompanying file "copying-gpl.txt" for more details.
 *
 * As a special exception to the GPL, permission is granted for additional
 * uses of the text contained in this file.  See the accompanying file
 * "copying-liquorice.txt" for details.
 */

/*
 * $Log: init.c,v $
 * Revision 1.17  2002/06/26 17:29:44  harald
 * First pre-release with 2.4 stack
 *
 */

#include <eeprom.h>

#include <sys/heap.h>
#include <sys/thread.h>
#include <sys/timer.h>

#include <sys/confos.h>

#ifdef NUTDEBUG
#include <sys/print.h>
#endif

CONFOS confos;

/*!
 * \addtogroup xgNutInit
 */
/*@{*/

extern void NutMain(void *arg) __attribute__ ((noreturn));

/*! \fn NutIdle(void *arg)
 * \brief Idle thread. 
 */
THREAD(NutIdle, arg)
{
    NutTimerInit();

    NutThreadCreate("main", NutMain, 0, 768);
    NutThreadSetPriority(255);
    for(;;) {
        NutThreadYield();
    }
}

/*!
 * \brief Load Nut/OS configuration from the EEPROM.
 *
 * This routine is automatically called during system
 * initialization.
 *
 * \return 0 if OK, -1 if configuration isn't available.
 */
int NutLoadConfig(void)
{
    eeprom_read_block(&confos, CONFOS_EE_OFFSET, sizeof(CONFOS));
    if(confos.size != sizeof(CONFOS) || confos.magic[0] != 'O' || confos.magic[1] != 'S') {
        return -1;
    }
    return 0;
}

/*!
 * \brief Save Nut/OS configuration in the EEPROM.
 *
 * \return 0 if OK, -1 on failures.
 */
int NutSaveConfig(void)
{
    u_char *cp;
    int i;

    confos.size = sizeof(CONFOS);
    confos.magic[0] = 'O';
    confos.magic[1] = 'S';
    for(cp = (u_char *)&confos, i = 0; i < sizeof(CONFOS); i++)
        eeprom_wb(i + CONFOS_EE_OFFSET, *cp++);

    return 0;
}

/*!
 * \brief Nut/OS main entry.
 *
 * This is the entry point for C programs. To hide neccessary
 * system initialization from the user program, this entry point
 * is included in the Nut/OS library as a separate object module.
 *
 * It will initialize memory management and the thread system
 * and start an idle thread, which in turn initializes the
 * timer functions. Finally NutMain() is called, which must
 * be defined by the application and may be used to start
 * additional application threads.
 * \internal
 */
int main(void)
{
#ifdef NUTDEBUG
    outp(1, UBRR);
    outp(BV(RXEN) | BV(TXEN), UCR);
#endif

    /*
     * Read eeprom configuration.
     */
    if(NutLoadConfig())
        NutSaveConfig();

    outp(BV(SRE) | BV(SRW), MCUCR);

    if((short)RAMEND - (short)(&__bss_end) > 384)
        NutHeapAdd(&__bss_end, (u_short)RAMEND - 256 - (u_short)(&__bss_end));
    NutHeapAdd((void *)(RAMEND + 1), 0x7FFF - RAMEND);
    
    NutThreadCreate("idle", NutIdle, 0, 384);

    return 0;
}

/*@}*/
