#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#ifdef _WIN32
#include <io.h>
#include "dirent.h"
#else
#include <dirent.h>
#endif

static int entryno = 0;

void dofile(char *name)
{
    int fd;
    unsigned char buf[512];
	int i;
	int cnt;
	long total = 0;

    fprintf(stderr, "Read %s\n", name);
    if((fd = open(name, O_RDONLY | O_BINARY)) == -1) {
        fprintf(stderr, "Failed to read file %s\n", name);
        return;
    }
	for(;;) {
		if((cnt = read(fd, buf, sizeof(buf))) < 0) {
			fprintf(stderr, "Failed reading file %s\n", name);
			break;
		}
		if(cnt == 0)
			break;
		if(total == 0) {
			entryno++;
			printf("char file%ddata[] __attribute__ ((progmem)) = {", entryno);
		}
		total += cnt;
		for(i = 0; i < cnt; i++) {
			if((i % 16) == 0)
				printf("\n");
			printf("0x%02x,", buf[i]);
		}
	}
    close(fd);
	if(total) {
		printf("\n};\nstatic ROMENTRY file%dentry = { ", entryno);
		if(entryno > 1)
			printf("&file%dentry, ", entryno - 1);
		else
			printf("0           , ", entryno - 1);
		if(name[0] == '.' && name[1] == '/')
			name += 2;	
		printf("\"%s\", %5d, file%ddata };\n", name, total, entryno);
	}
}

void dodir(char *dirpath)
{
    char path[256];
	DIR *dir;
	struct dirent *dire;
	struct stat statbuf;

    fprintf(stderr, "Scan %s\n", dirpath);
    if((dir = opendir(dirpath)) == NULL) {
        fprintf(stderr, "Failed to scan directory %s\n", dirpath);
        return;
    }
    while((dire = readdir(dir))) {
        if(dire->d_name[0] == '.' || stricmp(dire->d_name, "cvs") == 0)
            continue;
        strcpy(path, dirpath);
        strcat(path, "/");
        strcat(path, dire->d_name);
        stat(path, &statbuf);
  
        if(statbuf.st_mode & S_IFDIR)
            dodir(path);
        else if(statbuf.st_mode & S_IFREG)
            dofile(path);
    }
    closedir(dir);
}

int main(int argc, char *argv)
{
    printf("#include <fs/uromfs.h>\n\n");
    dodir(".");
	printf("\nROMENTRY *romEntryList = &file%dentry;\n", entryno);
    return 0;
}
