#include <stdio.h>
#include <windows.h>

extern int verbose;

/*!
 * \brief Install the GIVEIO port driver.
 */
int PortDriverInstall(u_char *path)
{
    SC_HANDLE manager;
    SC_HANDLE service;
    u_char *driver = "GIVEIO";

    if((manager = OpenSCManager(NULL, NULL,  SC_MANAGER_ALL_ACCESS)) == NULL) {
        return -1;
    }
    if((service = CreateService(manager, driver, driver, SERVICE_ALL_ACCESS, 
        SERVICE_KERNEL_DRIVER, SERVICE_AUTO_START, SERVICE_ERROR_NORMAL, path, 
        NULL, NULL, NULL, NULL, NULL)) == NULL) {
        CloseServiceHandle(manager);
        return -1;
    }
    CloseServiceHandle(service);
    CloseServiceHandle(manager);

    return 0;
}

/*!
 * \brief De-install the GIVEIO port driver.
 */
int PortDriverRemove(void)
{
    int rc = 0;
    SC_HANDLE manager;
    SC_HANDLE service;

    if((manager = OpenSCManager(NULL, NULL,  SC_MANAGER_ALL_ACCESS)) == NULL) {
        if(verbose)
            fprintf(stderr, "Error %u opening service manager\n", GetLastError());
        return -1;
    }
    if((service = OpenService(manager, "GIVEIO", SERVICE_ALL_ACCESS)) == NULL) {
        if(verbose)
            fprintf(stderr, "Error %u opening giveio service\n", GetLastError());
        CloseServiceHandle(manager);
        return -1;
    }
    if(!DeleteService(service)) {
        if(verbose)
            fprintf(stderr, "Error %u removing giveio service\n", GetLastError());
        rc = -1;
    }
    CloseServiceHandle(service);
    CloseServiceHandle(manager);

    return rc;
}
